# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""List key generator runs in a separate process to not block for the
potentially long key generation operation."""

import multiprocessing as mp

from flufl.lock import Lock
from pgpy import PGPKey, PGPUID
from pgpy.constants import (
    CompressionAlgorithm, HashAlgorithm, KeyFlags, SymmetricKeyAlgorithm)

from mailman_pgp.config import config
from mailman_pgp.utils.pgp import key_from_file


class ListKeyGenerator(mp.Process):
    """A multiprocessing list key generator."""

    def __init__(self, pgp_list):
        super().__init__(
                target=self._run,
                args=(config.pgp.primary_key_args, config.pgp.sub_key_args,
                      pgp_list.mlist.display_name,
                      pgp_list.mlist.posting_address,
                      pgp_list.mlist.request_address,
                      pgp_list.key_path),
                daemon=True)
        self._pgp_list = pgp_list

    def generate(self, block=False):
        self.start()
        if block:
            self.join()
            return key_from_file(self._pgp_list.key_path)

    def _run(self, primary_args, subkey_args, display_name, posting_address,
             request_address, key_path):
        """
        Generate the list keypair and save it.

        :param primary_args:
        :param subkey_args:
        :param display_name:
        :param posting_address:
        :param request_address:
        :param key_path:
        """
        self.key = self._create(primary_args, subkey_args, display_name,
                                posting_address,
                                request_address)
        with Lock(key_path + '.lock'):
            self._save(self.key, key_path)

    def _create(self, primary_args, subkey_args, display_name, posting_address,
                request_address):
        """
        Generate the list `PGPKey` keypair, with posting and request UIDs.

        Use a Sign+Certify main key and Encrypt subkey.
        :param primary_args:
        :param subkey_args:
        :param display_name:
        :param posting_address:
        :param request_address:
        :return: `PGPKey`
        """
        common_params = dict(
                hashes=[HashAlgorithm.SHA256,
                        HashAlgorithm.SHA384,
                        HashAlgorithm.SHA512,
                        HashAlgorithm.SHA224],
                ciphers=[SymmetricKeyAlgorithm.AES256,
                         SymmetricKeyAlgorithm.AES192,
                         SymmetricKeyAlgorithm.AES128],
                compression=[CompressionAlgorithm.ZLIB,
                             CompressionAlgorithm.BZ2,
                             CompressionAlgorithm.ZIP,
                             CompressionAlgorithm.Uncompressed]
        )

        # Generate the Sign + Certify primary key.
        key = PGPKey.new(*primary_args)
        key_params = dict(usage={KeyFlags.Sign, KeyFlags.Certify},
                          **common_params)
        # Generate the posting + request uids.
        main_uid = PGPUID.new(display_name, email=posting_address)
        request_uid = PGPUID.new(display_name,
                                 email=request_address)
        # Generate the Encrypt subkey.
        subkey = PGPKey.new(*subkey_args)
        subkey_params = dict(
                usage={KeyFlags.EncryptCommunications,
                       KeyFlags.EncryptStorage},
                **common_params
        )
        # Put it all together.
        key.add_uid(main_uid, primary=True, **key_params)
        key.add_uid(request_uid, **key_params)
        key.add_subkey(subkey, **subkey_params)
        return key

    def _save(self, key, key_path):
        """
        Save the generated key.

        :param key:
        :param key_path:
        """
        try:
            with open(key_path, 'w') as key_file:
                key_file.write(str(key))
        except FileNotFoundError:
            # Just eat it up.
            pass



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
""""""
from copy import deepcopy

from public import public


@public
class BaseWrapper:
    """"""

    def __init__(self, msg, copy=False):
        """
        Wrap the given message.

        :param msg: The message to wrap.
        :type msg: mailman.email.message.Message
        """
        if copy:
            self.msg = deepcopy(msg)
        else:
            self.msg = msg

    def copy(self):
        return self.__class__(self.msg, True)

    def __copy__(self):
        return self.copy()



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from public import public
from sqlalchemy import Integer, TypeDecorator


@public
class EnumFlag(TypeDecorator):
    """Simple bitwise flag from an `enum.Enum`, assumes it is unique."""
    impl = Integer

    def __init__(self, enum, *args, **kw):
        super().__init__(*args, **kw)
        self.enum = enum

    def process_bind_param(self, value, dialect):
        # value is a set/frozenset of Enums.
        if value is None:
            return None
        if not isinstance(value, (set, frozenset)):
            raise ValueError

        int_value = 0
        for enum in reversed(self.enum):
            int_value = int_value << 1
            if enum in value:
                int_value += 1

        return int_value

    def process_result_value(self, value, dialect):
        # value is an integer.
        if value is None:
            return None
        if value == 0:
            return frozenset()
        result = set()
        for enum in self.enum:
            if value % 2 == 1:
                result.add(enum)
            value = value >> 1

        return frozenset(result)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the module global PGP instance."""
import unittest

from mailman.app.lifecycle import create_list
from mailman.testing.helpers import subscribe

from mailman_pgp.config import config
from mailman_pgp.database import mm_transaction, transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.keygen import ListKeyGenerator
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key


class TestPGP(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
        self.pgp_list = PGPMailingList.for_list(self.mlist)
        self.list_key = ListKeyGenerator(self.pgp_list).generate(True)

        # Make Anne a member of this mailing list.
        self.anne = subscribe(self.mlist, 'Anne', email='anne@example.org')
        self.anne_key = load_key('rsa_1024.priv.asc')

        with transaction() as t:
            self.pgp_anne = PGPAddress(self.anne.address)
            self.pgp_anne.key = self.anne_key.pubkey
            self.pgp_anne.key_confirmed = True
            t.add(self.pgp_anne)

    def test_list_keydir(self):
        keyring = config.pgp.list_keyring
        self.assertEqual(len(keyring), 2)
        with keyring.key(self.pgp_list.mlist.fqdn_listname) as key:
            self.assertEqual(key.fingerprint, self.list_key.fingerprint)

    def test_user_keydir(self):
        keyring = config.pgp.user_keyring
        self.assertEqual(len(keyring), 2)
        with keyring.key(self.anne_key.fingerprint) as key:
            self.assertTrue(key.is_public)
            self.assertEqual(key.fingerprint, self.anne_key.fingerprint)

    def test_archive_keydir(self):
        keyring = config.pgp.archive_keyring
        self.assertEqual(len(keyring), 0)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from mailman.app.lifecycle import create_list

from mailman_pgp.database import mm_transaction
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.testing.layers import PGPConfigLayer


class TestDiscussionStyle(unittest.TestCase):
    layer = PGPConfigLayer

    def test_create(self):
        with mm_transaction():
            mlist = create_list('test@example.com', style_name='pgp-default')

        pgp_list = PGPMailingList.query().filter_by(
                list_id=mlist.list_id).first()

        # Test that we have our PGPMailingList
        self.assertIsNotNone(pgp_list)
        self.assertEqual(pgp_list.mlist, mlist)

        # from LegacyDiscussionStyle
        self.assertEqual(mlist.allow_list_posts, True)
        self.assertEqual(mlist.send_welcome_message, True)
        self.assertEqual(mlist.send_goodbye_message, True)
        self.assertEqual(mlist.anonymous_list, False)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Signature checking rule for the pgp-posting-chain."""
from operator import attrgetter

from mailman.core.i18n import _
from mailman.interfaces.action import Action
from mailman.interfaces.chain import AcceptEvent
from mailman.interfaces.rules import IRule
from mailman.interfaces.usermanager import IUserManager
from public import public
from zope.component import getUtility
from zope.event import classhandler
from zope.interface import implementer

from mailman_pgp.config import config
from mailman_pgp.database import transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.model.sighash import PGPSigHash
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.utils.email import get_email
from mailman_pgp.utils.moderation import record_action
from mailman_pgp.utils.pgp import expired, hashes, revoked, verifies


@public
@implementer(IRule)
class Signature:
    """The signature checking rule."""

    name = 'pgp-signature'
    description = _(
            'A rule which enforces PGP enabled list signature configuration.')
    record = False

    def check(self, mlist, msg, msgdata):
        """See `IRule`."""
        # Find the `PGPMailingList` this is for.
        pgp_list = PGPMailingList.for_list(mlist)
        if pgp_list is None:
            return False

        email = get_email(msg)
        # Wrap the message to work with it.
        wrapped = PGPWrapper(msg)

        # Take unsigned_msg_action if unsigned.
        if not wrapped.is_signed():
            action = pgp_list.unsigned_msg_action
            if action != Action.defer:
                record_action(msg, msgdata, action, email,
                              'The message is unsigned.')
                return True

        # Take `inline_pgp_action` if inline signed.
        if wrapped.inline.is_signed():
            action = pgp_list.inline_pgp_action
            if action != Action.defer:
                record_action(msg, msgdata, action, email,
                              'Inline PGP is not allowed.')
                return True

        # Lookup the address by sender, and its corresponding `PGPAddress`.
        user_manager = getUtility(IUserManager)
        address = user_manager.get_address(email)
        pgp_address = PGPAddress.for_address(address)
        if pgp_address is None:
            # Just let it continue.
            return False

        # See if we have a key.
        key = pgp_address.key
        if key is None:
            record_action(msg, msgdata, Action.reject, email,
                          'No key set for address {}.'.format(email))
            return True

        if not pgp_address.key_confirmed:
            record_action(msg, msgdata, Action.reject, email,
                          'Key not confirmed.')
            return True

        verifications = list(wrapped.verify(key))
        # verifications is a list of SignatureVerification, only contains
        # sigs that appear to be by the pgp_address.key

        if expired(verifications):
            action = pgp_list.expired_sig_action
            if action != Action.defer:
                record_action(msg, msgdata, action, email,
                              'Signature is expired.')
                return True

        if revoked(verifications):
            action = pgp_list.revoked_sig_action
            if action != Action.defer:
                record_action(msg, msgdata, action, email,
                              'Signature is made by a revoked key.')
                return True

        # Take the `invalid_sig_action` if the verification failed.
        if not verifies(verifications):
            action = pgp_list.invalid_sig_action
            if action != Action.defer:
                record_action(msg, msgdata, action, email,
                              'Signature did not verify.')
                return True

        sig_hashes = set(hashes(verifications))
        duplicates = set(PGPSigHash.hashes(sig_hashes))
        if duplicates:
            fingerprints = map(attrgetter('fingerprint'), duplicates)
            if key.fingerprint in fingerprints:
                action = pgp_list.duplicate_sig_action
                if action != Action.defer:
                    record_action(msg, msgdata, action, email,
                                  'Signature duplicate.')
                    return True
        msgdata['pgp_sig_hashes'] = sig_hashes

        return False


@classhandler.handler(AcceptEvent)
def on_message_posting(event):
    """
    Add sig hashes to sighash table.

    :param event:
    :type event: AcceptEvent
    """
    if not config.get_value('misc', 'collect_sig_hashes'):
        return
    pgp_list = PGPMailingList.for_list(event.mlist)
    if pgp_list is None:
        return
    address = getUtility(IUserManager).get_address(get_email(event.msg))
    pgp_address = PGPAddress.for_address(address)
    if pgp_address is None or pgp_address.key_fingerprint is None:
        return
    for sig_hash in event.msgdata['pgp_sig_hashes']:
        with transaction() as t:
            pgp_hash = PGPSigHash(hash=sig_hash,
                                  fingerprint=pgp_address.key_fingerprint)
            t.add(pgp_hash)



import sys

from setuptools import find_packages, setup

if sys.hexversion < 0x30500f0:
    print('Mailman requires at least Python 3.5')
    sys.exit(1)

from importlib.machinery import SourceFileLoader

_init = SourceFileLoader('__init__',
                         'src/mailman_pgp/__init__.py').load_module()

setup(
        name='mailman_pgp',
        version=_init.__version__,
        description='A PGP plugin for the GNU Mailman mailing list manager',
        long_description="""\
A plugin for GNU Mailman that adds encrypted mailing lists via PGP/MIME.""",
        url='https://gitlab.com/J08nY/mailman-pgp',
        author=_init.__author__,
        author_email='johny@neuromancer.sk',
        license=_init.__license__,
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'Intended Audience :: System Administrators',
            'Operating System :: POSIX',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Topic :: Communications :: Email :: Mailing List Servers',
        ],
        keywords='email pgp',
        packages=find_packages('src'),
        package_dir={'': 'src'},
        include_package_data=True,
        zip_safe=False,
        python_requires='>=3.5',
        install_requires=[
            'mailman>=3.2.0a1',
            'PGPy>=0.4.2',
            'atpublic',
            'flufl.lock',
            'sqlalchemy',
            'zope.interface',
            'zope.event'
        ],
        tests_require=[
            'flufl.testing',
            'parameterized',
            'nose2'
        ],
        test_suite='nose2.collector.collector',
        dependency_links=[
            'https://github.com/J08nY/PGPy/archive/dev.zip#egg=PGPy-0.4.2',
            'https://gitlab.com/J08nY/mailman/repository/archive.tar.gz?ref=plugin#egg=mailman-3.2.0a1'
        ]
)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
from mailman.interfaces.rules import IRule
from public import public
from zope.interface import implementer


@public
@implementer(IRule)
class MarkPosting:
    """"""

    name = 'pgp-mark'
    description = ''
    record = False

    def check(self, mlist, msg, msgdata):
        """See `IRule`."""
        msgdata['pgp_is_posting'] = True
        return False



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""

from mailman.styles.default import LegacyAnnounceOnly
from public import public

from mailman_pgp.styles.base import PGPStyle


@public
class AnnounceStyle(LegacyAnnounceOnly, PGPStyle):
    name = 'pgp-announce'
    description = 'Announce only PGP enabled mailing list style.'

    def apply(self, mailing_list):
        """See `IStyle`."""
        LegacyAnnounceOnly.apply(self, mailing_list)
        PGPStyle.apply(self, mailing_list)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Archives messages by sending to django-pgpmailman,
an extension on top of Postorius and HyperKitty.
"""

from mailman.interfaces.archiver import IArchiver
from public import public
from zope.interface import implementer


@public
@implementer(IArchiver)
class RemoteArchiver:
    """Remote PGP-enabled archiver."""
    pass



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the PGP posting chain."""
import unittest

from mailman.app.lifecycle import create_list
from mailman.email.message import Message

from mailman_pgp.chains.default import PGPChain
from mailman_pgp.testing.layers import PGPConfigLayer


class TestPGPChain(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.mlist = create_list('test@example.com', style_name='pgp-default')

    def test_cached_links(self):
        chain = PGPChain()
        links = list(chain.get_links(self.mlist, Message(), {}))
        cached_links = list(chain.get_links(self.mlist, Message(), {}))
        for link, cached in zip(links, cached_links):
            self.assertIs(link, cached)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import time
import unittest
from datetime import timedelta

from mailman.app.lifecycle import create_list
from mailman.email.message import Message
from mailman.interfaces.action import Action
from mailman.interfaces.chain import AcceptEvent
from mailman.interfaces.member import MemberRole
from mailman.interfaces.usermanager import IUserManager
from mailman.testing.helpers import (set_preferred,
                                     specialized_message_from_string as mfs)
from zope.component import getUtility
from zope.event import notify

from mailman_pgp.chains.default import PGPChain
from mailman_pgp.config import mm_config
from mailman_pgp.database import mm_transaction, transaction
from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.model.sighash import PGPSigHash
from mailman_pgp.pgp.inline import InlineWrapper
from mailman_pgp.pgp.mime import MIMEWrapper
from mailman_pgp.pgp.wrapper import PGPWrapper
from mailman_pgp.rules.signature import Signature
from mailman_pgp.testing.config import patch_config
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key, load_message
from mailman_pgp.utils.pgp import hashes


class TestPGPSignatureRule(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.rule = Signature()

        user_manager = getUtility(IUserManager)
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
            self.sender = user_manager.create_user('RSA-1024b@example.org')
            set_preferred(self.sender)
            self.mlist.subscribe(self.sender, MemberRole.member)

        self.pgp_list = PGPMailingList.for_list(self.mlist)

        self.sender_key = load_key('rsa_1024.priv.asc')
        with transaction() as t:
            self.pgp_sender = PGPAddress(self.sender.preferred_address)
            self.pgp_sender.key = self.sender_key.pubkey
            self.pgp_sender.key_confirmed = True
            t.add(self.pgp_sender)

        self.msg_clear = load_message('clear.eml')
        self.msg_inline_signed = load_message('inline_signed.eml')
        self.msg_mime_signed = load_message('mime_signed.eml')
        self.msg_inline_signed_invalid = load_message(
                'inline_cleartext_signed_invalid.eml')
        self.msg_mime_signed_invalid = load_message(
                'mime_signed_invalid.eml')

    def assertAction(self, msgdata, action, reasons):
        self.assertEqual(msgdata['pgp_action'], action.name)
        self.assertListEqual(msgdata['moderation_reasons'], reasons)

    def test_has_rule(self):
        self.assertIn(Signature.name, mm_config.rules.keys())

    def test_no_pgp_list(self):
        with mm_transaction():
            ordinary_list = create_list('odrinary@example.com')
        msg = mfs("""\
From: anne@example.com
To: ordinary@example.com

""")

        matches = self.rule.check(ordinary_list, msg, {})
        self.assertFalse(matches)

    def test_no_address(self):
        with transaction():
            self.pgp_list.unsigned_msg_action = Action.defer
        msg = mfs("""\
From: anne@example.com
To: test@example.com

""")
        matches = self.rule.check(self.mlist, msg, {})
        self.assertFalse(matches)

    def test_no_key(self):
        with transaction():
            self.pgp_sender.key = None

        msgdata = {}
        matches = self.rule.check(self.mlist, self.msg_mime_signed, msgdata)
        self.assertTrue(matches)
        self.assertAction(msgdata, Action.reject, [
            'No key set for address {}.'.format(
                    self.pgp_sender.address.original_email)])

    def test_key_not_confirmed(self):
        with transaction():
            self.pgp_sender.key_confirmed = False

        msgdata = {}
        matches = self.rule.check(self.mlist, self.msg_mime_signed, msgdata)
        self.assertTrue(matches)
        self.assertAction(msgdata, Action.reject, ['Key not confirmed.'])

    def test_unsigned_action(self):
        with transaction():
            self.pgp_list.unsigned_msg_action = Action.hold
            self.pgp_list.inline_pgp_action = Action.defer
            self.pgp_list.expired_sig_action = Action.defer
            self.pgp_list.invalid_sig_action = Action.defer
            self.pgp_list.revoked_sig_action = Action.defer
            self.pgp_list.duplicate_sig_action = Action.defer

        msgdata = {}
        matches = self.rule.check(self.mlist, self.msg_clear, msgdata)
        self.assertTrue(matches)
        self.assertAction(msgdata, Action.hold, ['The message is unsigned.'])

        matches = self.rule.check(self.mlist, self.msg_inline_signed, msgdata)
        self.assertFalse(matches)

        matches = self.rule.check(self.mlist, self.msg_mime_signed, msgdata)
        self.assertFalse(matches)

    def test_inline_pgp_action(self):
        with transaction():
            self.pgp_list.unsigned_msg_action = Action.defer
            self.pgp_list.inline_pgp_action = Action.hold
            self.pgp_list.expired_sig_action = Action.defer
            self.pgp_list.invalid_sig_action = Action.defer
            self.pgp_list.revoked_sig_action = Action.defer
            self.pgp_list.duplicate_sig_action = Action.defer

        msgdata = {}
        matches = self.rule.check(self.mlist, self.msg_inline_signed, msgdata)
        self.assertTrue(matches)
        self.assertAction(msgdata, Action.hold, ['Inline PGP is not allowed.'])

        matches = self.rule.check(self.mlist, self.msg_mime_signed, msgdata)
        self.assertFalse(matches)

    def test_expired_sig_action(self):
        with transaction():
            self.pgp_list.unsigned_msg_action = Action.defer
            self.pgp_list.inline_pgp_action = Action.defer
            self.pgp_list.expired_sig_action = Action.hold
            self.pgp_list.invalid_sig_action = Action.defer
            self.pgp_list.revoked_sig_action = Action.defer
            self.pgp_list.duplicate_sig_action = Action.defer

        msgdata = {}
        wrapped = MIMEWrapper(self.msg_clear, True)
        msg = wrapped.sign(self.sender_key, expires=timedelta(seconds=1)).msg
        time.sleep(2)
        matches = self.rule.check(self.mlist, msg, msgdata)

        self.assertTrue(matches)
        self.assertAction(msgdata, Action.hold, ['Signature is expired.'])

        msgdata = {}
        wrapped = InlineWrapper(self.msg_clear, True)
        msg = wrapped.sign(self.sender_key, expires=timedelta(seconds=1)).msg
        time.sleep(2)
        matches = self.rule.check(self.mlist, msg, msgdata)

        # TODO: test when the key is expired

        self.assertTrue(matches)
        self.assertAction(msgdata, Action.hold, ['Signature is expired.'])

    def test_revoked_sig_action(self):
        with transaction():
            self.pgp_list.unsigned_msg_action = Action.defer
            self.pgp_list.inline_pgp_action = Action.defer
            self.pgp_list.expired_sig_action = Action.defer
            self.pgp_list.revoked_sig_action = Action.hold
            self.pgp_list.invalid_sig_action = Action.defer
            self.pgp_list.duplicate_sig_action = Action.defer

            rsig = self.sender_key.revoke(self.pgp_sender.key)
            self.pgp_sender.key |= rsig

        msgdata = {}
        wrapped = MIMEWrapper(self.msg_clear)
        msg = wrapped.sign(self.sender_key).msg
        matches = self.rule.check(self.mlist, msg, msgdata)

        self.assertTrue(matches)
        self.assertAction(msgdata, Action.hold,
                          ['Signature is made by a revoked key.'])

    def test_invalid_sig_action(self):
        with transaction():
            self.pgp_list.unsigned_msg_action = Action.defer
            self.pgp_list.inline_pgp_action = Action.defer
            self.pgp_list.expired_sig_action = Action.defer
            self.pgp_list.invalid_sig_action = Action.hold
            self.pgp_list.revoked_sig_action = Action.defer
            self.pgp_list.duplicate_sig_action = Action.defer

        msgdata = {}
        matches = self.rule.check(self.mlist, self.msg_inline_signed_invalid,
                                  msgdata)
        self.assertTrue(matches)
        self.assertAction(msgdata, Action.hold, ['Signature did not verify.'])

        msgdata = {}
        matches = self.rule.check(self.mlist, self.msg_mime_signed_invalid,
                                  msgdata)
        self.assertTrue(matches)
        self.assertAction(msgdata, Action.hold, ['Signature did not verify.'])

    def test_duplicate_sig_action(self):
        with transaction() as t:
            self.pgp_list.unsigned_msg_action = Action.defer
            self.pgp_list.inline_pgp_action = Action.defer
            self.pgp_list.expired_sig_action = Action.defer
            self.pgp_list.invalid_sig_action = Action.defer
            self.pgp_list.revoked_sig_action = Action.defer
            self.pgp_list.duplicate_sig_action = Action.hold

            wrapped = PGPWrapper(self.msg_mime_signed)
            sig_hashes = set(hashes(wrapped.verify(self.sender_key.pubkey)))
            wrapped = PGPWrapper(self.msg_inline_signed)
            sig_hashes |= set(hashes(wrapped.verify(self.sender_key.pubkey)))
            for hash in sig_hashes:
                sig_hash = PGPSigHash()
                sig_hash.hash = hash
                sig_hash.fingerprint = self.sender_key.pubkey.fingerprint
                t.add(sig_hash)

        msgdata = {}
        matches = self.rule.check(self.mlist, self.msg_mime_signed, msgdata)
        self.assertTrue(matches)
        self.assertAction(msgdata, Action.hold, ['Signature duplicate.'])

        msgdata = {}
        matches = self.rule.check(self.mlist, self.msg_inline_signed, msgdata)
        self.assertTrue(matches)
        self.assertAction(msgdata, Action.hold, ['Signature duplicate.'])


class TestPostingEvent(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        self.rule = Signature()

        user_manager = getUtility(IUserManager)
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
            self.sender = user_manager.create_user('RSA-1024b@example.org')
            set_preferred(self.sender)
            self.mlist.subscribe(self.sender, MemberRole.member)

        self.pgp_list = PGPMailingList.for_list(self.mlist)

        self.sender_key = load_key('rsa_1024.priv.asc')
        with transaction() as t:
            self.pgp_sender = PGPAddress(self.sender.preferred_address)
            self.pgp_sender.key = self.sender_key.pubkey
            self.pgp_sender.key_confirmed = True
            t.add(self.pgp_sender)

    def test_sighashes_added(self):
        msg = load_message('mime_signed.eml')
        wrapped = PGPWrapper(msg)
        sighashes = set(hashes(wrapped.verify(self.sender_key)))
        msgdata = dict(pgp_sig_hashes=sighashes)
        notify(AcceptEvent(self.mlist, msg, msgdata,
                           mm_config.chains[PGPChain.name]))

        for hash in sighashes:
            sig_hash = PGPSigHash.query().filter_by(hash=hash).one()
            self.assertIsNotNone(sig_hash)
            self.assertEqual(sig_hash.fingerprint, self.sender_key.fingerprint)

    @patch_config('misc', 'collect_sig_hashes', 'no')
    def test_no_collect(self):
        msg = load_message('mime_signed.eml')
        wrapped = PGPWrapper(msg)
        sighashes = set(hashes(wrapped.verify(self.sender_key)))
        msgdata = dict(pgp_sig_hashes=sighashes)
        notify(AcceptEvent(self.mlist, msg, msgdata,
                           mm_config.chains[PGPChain.name]))

        self.assertEqual(0, len(PGPSigHash.query().all()))

    def test_no_pgp_list(self):
        with mm_transaction():
            mlist = create_list('ordinary@example.com')
        notify(AcceptEvent(mlist, Message(), dict(),
                           mm_config.chains[PGPChain.name]))

    def test_no_pgp_address(self):
        msg = mfs("""\
From: anne@example.com
To: test@example.com
Subject: something

Some text.
""")
        notify(AcceptEvent(self.mlist, msg, dict(),
                           mm_config.chains[PGPChain.name]))



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""PGP enabled BulkDelivery."""
import copy

from mailman.mta.bulk import BulkDelivery
from public import public

from mailman_pgp.model.address import PGPAddress
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.mime import MIMEWrapper


class CallbackBulkDelivery(BulkDelivery):
    """Bulk delivery that has a list of callbacks to run for each chunk."""

    def __init__(self, max_recipients=None):
        super().__init__(max_recipients=max_recipients)
        self.callbacks = []

    def deliver(self, mlist, msg, msgdata):
        """See `IMailTransportAgentDelivery`."""
        refused = {}
        for recipients in self.chunkify(msgdata.get('recipients', set())):
            message_copy = copy.deepcopy(msg)
            msgdata_copy = msgdata.copy()
            recipients_copy = set(recipients)

            for callback in self.callbacks:
                callback(mlist, message_copy, msgdata_copy, recipients_copy)
            callback_refused = dict(
                    (recipient, (444, BaseException))
                    for recipient in recipients - recipients_copy)
            refused.update(callback_refused)

            chunk_refused = self._deliver_to_recipients(
                    mlist, message_copy, msgdata_copy, recipients_copy)
            refused.update(chunk_refused)
        return refused


class PGPBulkMixin:
    """Bulk encryption and signing Delivery mixin."""

    def sign_encrypt(self, mlist, msg, msgdata, recipients):
        """
        Sign and encrypt the outgoing message to the recipients.

        :param mlist:
        :type mlist: mailman.model.mailinglist.MailingList
        :param msg:
        :type msg: mailman.email.message.Message
        :param msgdata:
        :type msgdata: dict
        :param recipients:
        :type recipients: set
        """
        pgp_list = PGPMailingList.for_list(mlist)
        if not pgp_list:
            return
        if not pgp_list.encrypt_outgoing and not pgp_list.sign_outgoing:
            # nothing to do
            return

        keys = []
        for recipient in set(recipients):
            pgp_address = PGPAddress.for_email(recipient)
            if pgp_address is None:
                recipients.remove(recipient)
                continue
            if pgp_address.key is None or not pgp_address.key_confirmed:
                recipients.remove(recipient)
                continue
            keys.append(pgp_address.key)

        wrapped = MIMEWrapper(msg)
        if pgp_list.sign_outgoing:
            if pgp_list.encrypt_outgoing:
                wrapped.sign_encrypt(pgp_list.key, pgp_list.pubkey,
                                     *keys, throw_keyid=True)
            else:
                wrapped.sign(pgp_list.key)
        else:
            # Definitely encrypt here, the case where we don't encrypt or sign
            # is handled above at the start of the func.
            wrapped.encrypt(pgp_list.pubkey, *keys, throw_keyid=True)


@public
class PGPBulkDelivery(CallbackBulkDelivery, PGPBulkMixin):
    """Bulk PGP enabled delivery."""

    def __init__(self, max_recipients=None):
        super().__init__(max_recipients=max_recipients)
        self.callbacks.append(self.sign_encrypt)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Strict inline PGP message wrapper."""
import copy
from email.iterators import walk
from email.mime.text import MIMEText

from pgpy import PGPMessage
from pgpy.constants import SymmetricKeyAlgorithm
from public import public

from mailman_pgp.pgp.base import BaseWrapper
from mailman_pgp.utils.pgp import key_from_blob, revoc_from_blob


@public
class InlineWrapper(BaseWrapper):
    """Inline PGP wrapper."""

    def get_payload(self):
        for part in walk(self.msg):
            if not part.is_multipart():
                yield part.get_payload()

    def _walk(self, walk_fn, *args, **kwargs):
        for part in walk(self.msg):
            if not part.is_multipart():
                yield walk_fn(part, *args, **kwargs)

    def _is_signed(self, part):
        try:
            msg = PGPMessage.from_blob(part.get_payload())
            return msg.is_signed
        except:
            pass
        return False

    def is_signed(self):
        """
        Whether the message is inline signed.

        :return: If the message is inline signed.
        :rtype: bool
        """
        return all(self._walk(self._is_signed))

    def has_signature(self):
        """
        Whether some parts of the message are inline signed.

        :return: If some parts of the message are inline signed.
        :rtype: bool
        """
        return any(self._walk(self._is_signed))

    def get_signed(self):
        """

        :return:
        """
        for part in walk(self.msg):
            if not part.is_multipart() and self._is_signed(part):
                try:
                    msg = PGPMessage.from_blob(part.get_payload()).message
                except:
                    continue
                yield msg

    def get_signature(self):
        """

        :return:
        :rtype: typing.Generator[pgpy.PGPMessage]
        """
        for part in walk(self.msg):
            if not part.is_multipart():
                try:
                    msg = PGPMessage.from_blob(part.get_payload())
                except:
                    continue
                yield msg

    def strip_signature(self):
        for part in walk(self.msg):
            if not part.is_multipart() and self._is_signed(part):
                try:
                    msg = PGPMessage.from_blob(part.get_payload())
                except:
                    continue
                part.set_payload(msg.message)
        return self

    def _is_encrypted(self, part):
        try:
            msg = PGPMessage.from_blob(part.get_payload())
            return msg.is_encrypted
        except:
            pass
        return False

    def is_encrypted(self):
        """
        Whether the message is inline encrypted.

        :return: If the message is inline encrypted.
        :rtype: bool
        """
        return all(self._walk(self._is_encrypted))

    def has_encryption(self):
        """
        Whether some parts of the message are inline encrypted.

        :return: If some parts of the message are inline encrypted.
        :rtype: bool
        """
        return any(self._walk(self._is_encrypted))

    def get_encrypted(self):
        """

        :return:
        :rtype: typing.Generator[pgpy.PGPMessage]
        """
        for part in walk(self.msg):
            if not part.is_multipart():
                try:
                    msg = PGPMessage.from_blob(part.get_payload())
                except:
                    continue
                yield msg

    def _has_keys(self, part):
        try:
            key_from_blob(part.get_payload())
            return True
        except:
            pass
        return False

    def is_keys(self):
        """
        Whether the message is all keys (all parts).

        :return: If the message is keys.
        :rtype: bool
        """
        return all(self._walk(self._has_keys))

    def has_keys(self):
        """
        Whether the message contains public or private keys.

        :return: If the message contains keys.
        :rtype: bool
        """
        return any(self._walk(self._has_keys))

    def keys(self):
        """
        Get the collection of keys in this message.

        :return: A collection of keys.
        :rtype: Generator[pgpy.PGPKey]
        """
        for part in walk(self.msg):
            if not part.is_multipart():
                try:
                    key = key_from_blob(part.get_payload())
                except:
                    continue
                yield key

    def _is_revoc(self, part):
        try:
            revoc_from_blob(part.get_payload())
        except ValueError:
            return False
        return True

    def is_revocs(self):
        for part in walk(self.msg):
            if (not part.is_multipart() and not self._is_revoc(part)):
                return False
        return True

    def has_revocs(self):
        for part in walk(self.msg):
            if (not part.is_multipart() and self._is_revoc(part)):
                return True
        return False

    def revocs(self):
        for part in walk(self.msg):
            if not part.is_multipart():
                try:
                    revoc = revoc_from_blob(part.get_payload())
                except:
                    continue
                yield revoc

    def attach_revocs(self, *key_revocations):
        """
        Attach a key revocation signature to the message.

        :param key_revocations: A key revocation signature to attach.
        :type key_revocations: pgpy.PGPSignature
        :return:
        :rtype: InlineWrapper
        """
        if self.msg.get_content_type() != 'multipart/mixed':
            # wrap in multipart/mixed
            payload = copy.deepcopy(self.msg)
            self.msg.set_payload([])
            self.msg.set_type('multipart/mixed')
            self.msg['MIME-Version'] = '1.0'
            self.msg.attach(payload)

        for key_revocation in key_revocations:
            revoc_part = MIMEText(str(key_revocation))
            self.msg.attach(revoc_part)
        return self

    def verify(self, key):
        """
        Verify the signatures of this message with key.

        :param key: The key to verify with.
        :type key: pgpy.PGPKey
        :return: The verified signatures.
        :rtype: Generator[pgpy.types.SignatureVerification]
        """
        yield from map(key.verify, self.get_signature())

    def _sign(self, pmsg, key, **kwargs):
        smsg = copy.copy(pmsg)
        smsg |= key.sign(smsg, **kwargs)
        return smsg

    def sign(self, key, **kwargs):
        """
        Sign a message with key.

        :param key: The key to sign with.
        :type key: pgpy.PGPKey
        :return:
        :rtype: InlineWrapper
        """
        for part in walk(self.msg):
            if not part.is_multipart():
                if self._is_signed(part):
                    pmsg = PGPMessage.from_blob(part.get_payload())
                else:
                    payload = str(part.get_payload())
                    pmsg = PGPMessage.new(payload, cleartext=True)
                smsg = self._sign(pmsg, key, **kwargs)
                part.set_payload(str(smsg))
        return self

    def _decrypt(self, part, key):
        message = PGPMessage.from_blob(part.get_payload())
        decrypted = key.decrypt(message)
        if decrypted.is_signed:
            part.set_payload(str(decrypted))
        else:
            dmsg = decrypted.message
            if isinstance(dmsg, bytearray):
                dmsg = dmsg.decode(decrypted.charset or 'utf-8')
            part.set_payload(dmsg)

    def decrypt(self, key):
        """
        Decrypt this message with key.

        :param key: The key to decrypt with.
        :type key: pgpy.PGPKey
        :return:
        :rtype: InlineWrapper
        """
        for part in walk(self.msg):
            if not part.is_multipart() and self._is_encrypted(part):
                self._decrypt(part, key)
        return self

    def _encrypt(self, pmsg, *keys, cipher, **kwargs):
        emsg = copy.copy(pmsg)
        if len(keys) == 1:
            emsg = keys[0].encrypt(emsg, cipher=cipher, **kwargs)
        else:
            session_key = cipher.gen_key()
            for key in keys:
                emsg = key.encrypt(emsg, cipher=cipher,
                                   sessionkey=session_key,
                                   **kwargs)
            del session_key
        return emsg

    def encrypt(self, *keys, cipher=SymmetricKeyAlgorithm.AES256,
                **kwargs):
        """
        Encrypt the message with key/s, using cipher.

        :param keys: The key/s to encrypt with.
        :type keys: pgpy.PGPKey
        :param cipher: The symmetric cipher to use.
        :type cipher: SymmetricKeyAlgorithm
        :return:
        :rtype: InlineWrapper
        """
        if len(keys) == 0:
            raise ValueError('At least one key necessary.')

        for part in walk(self.msg):
            if not part.is_multipart():
                payload = str(part.get_payload())
                pmsg = PGPMessage.new(payload)
                emsg = self._encrypt(pmsg, *keys, cipher=cipher, **kwargs)
                part.set_payload(str(emsg))
        return self

    def sign_encrypt(self, key, *keys, hash=None,
                     cipher=SymmetricKeyAlgorithm.AES256,
                     **kwargs):
        """
        Sign and encrypt the message, in one go.

        :param key: The key to sign with.
        :type key: pgpy.PGPKey
        :param keys: The key/s to encrypt with.
        :type keys: pgpy.PGPKey
        :param hash:
        :type hash: pgpy.constants.HashAlgorithm
        :param cipher:
        :type cipher: pgpy.constants.SymmetricKeyAlgorithm
        :return:
        :rtype: InlineWrapper
        """
        if len(keys) == 0:
            raise ValueError('At least one key necessary.')

        for part in walk(self.msg):
            if not part.is_multipart():
                if self._is_signed(part):
                    pmsg = PGPMessage.from_blob(part.get_payload())
                else:
                    payload = str(part.get_payload())
                    pmsg = PGPMessage.new(payload)
                smsg = self._sign(pmsg, key, hash=hash)
                emsg = self._encrypt(smsg, *keys, cipher=cipher, **kwargs)
                part.set_payload(str(emsg))
        return self



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""

from mailman.rest.helpers import (child, CollectionMixin, etag, not_found,
                                  NotFound, okay)
from public.public import public

from mailman_pgp.config import config
from mailman_pgp.model.address import PGPAddress


class _PGPAddressBase(CollectionMixin):
    def _resource_as_dict(self, address):
        """See `CollectionMixin`."""
        return dict(email=address.email,
                    key_fingerprint=address.key_fingerprint,
                    key_confirmed=address.key_confirmed,
                    self_link=self.api.path_to(
                            '/plugins/{}/addresses/{}'.format(config.name,
                                                              address.email)
                    ))

    def _get_collection(self, request):
        """See `CollectionMixin`."""
        return PGPAddress.query().all()


@public
class AllAddresses(_PGPAddressBase):
    """"""

    def on_get(self, request, response):
        """/addresses"""
        resource = self._make_collection(request)
        return okay(response, etag(resource))


@public
class AnAddress(_PGPAddressBase):
    """"""

    def __init__(self, email):
        self._address = PGPAddress.for_email(email)

    def on_get(self, request, response):
        if self._address is None:
            return not_found(response)
        else:
            okay(response, self._resource_as_json(self._address))

    @child()
    def key(self, context, segments):
        if self._address is None:
            return NotFound(), []
        return AnAddressKey(self._address), []


@public
class AnAddressKey:
    """"""

    def __init__(self, address):
        self._address = address

    def on_get(self, request, response):
        """/addresses/<email>/key"""
        key = self._address.key
        if key is None:
            not_found(response)
        else:
            resource = dict(key=str(key),
                            key_fingerprint=str(key.fingerprint),
                            key_confirmed=self._address.key_confirmed)
            okay(response, etag(resource))



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import shutil
import unittest
from os.path import exists

from mailman.app.lifecycle import create_list
from mailman.interfaces.listmanager import IListManager
from zope.component import getUtility

from mailman_pgp.database import mm_transaction
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.testing.config import patch_config
from mailman_pgp.testing.layers import PGPConfigLayer
from mailman_pgp.testing.pgp import load_key


class TestPGPMailingList(unittest.TestCase):
    layer = PGPConfigLayer

    def setUp(self):
        with mm_transaction():
            self.mlist = create_list('test@example.com',
                                     style_name='pgp-default')
        pgp_list = PGPMailingList.for_list(self.mlist)
        pgp_list.key = load_key('rsa_1024.priv.asc')

    def test_delete(self):
        getUtility(IListManager).delete(self.mlist)
        pgp_list = PGPMailingList.for_list(self.mlist)
        self.assertIsNone(pgp_list)

    def test_shred_key(self):
        key_path = PGPMailingList.for_list(self.mlist).key_path
        getUtility(IListManager).delete(self.mlist)
        self.assertFalse(exists(key_path))

    @patch_config('keypairs', 'shred_command', 'shred')
    @unittest.skipIf(shutil.which('shred') is None, 'No shred command.')
    def test_shred_key_command(self):
        # This really just hopes that the test env has some coreutils, that
        # provide `shred` command with this behaviour.
        key_path = PGPMailingList.for_list(self.mlist).key_path
        with open(key_path, 'rb') as f:
            before = f.read()
        getUtility(IListManager).delete(self.mlist)
        with open(key_path, 'rb') as f:
            after = f.read()
        self.assertNotEqual(before, after)

    @patch_config('keypairs', 'shred', 'no')
    def test_delete_key(self):
        key_path = PGPMailingList.for_list(self.mlist).key_path
        getUtility(IListManager).delete(self.mlist)
        self.assertFalse(exists(key_path))



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Config validation."""
from configparser import ConfigParser


class ConfigValidator:
    """Validates a ConfigParser object against a schema."""

    def __init__(self, schema):
        self.schema = ConfigParser()
        self.schema.read_string(schema)

    def validate(self, cfg):
        """

        :param cfg:
        :type cfg: ConfigParser
        """
        errors = []
        additional_sections = set(cfg.sections()).difference(
                self.schema.sections())

        if len(additional_sections) > 0:
            errors.append(
                    'Additional sections: {}'.format(additional_sections))

        for section in self.schema.sections():
            if not cfg.has_section(section):
                errors.append('Missing config section: {}'.format(section))
                continue
            for option in self.schema.options(section):
                if not cfg.has_option(section, option):
                    errors.append(
                            'Missing config option {} in {}'.format(option,
                                                                    section))
            additional_options = set(cfg.options(section)).difference(
                    self.schema.options(section))
            if len(additional_options) > 0:
                errors.append(
                        'Additional options: {}'.format(additional_options))

        if len(errors) > 0:
            raise ValueError(errors)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

"""Config conversion."""
import builtins
import re
from configparser import ConfigParser

from mailman.utilities.modules import find_name


class ConfigConverter:
    """Converts a ConfigParser object values according to a schema."""

    def __init__(self, schema):
        self.schema = ConfigParser()
        self.schema.read_string(schema)

    def convert(self, cfg):
        """

        :param cfg:
        :type cfg: ConfigParser
        """
        out = dict()
        for section in self.schema.sections():
            out[section] = dict()
            for option in self.schema.options(section):
                out[section][option] = self._transform_option(section, option,
                                                              cfg.get(section,
                                                                      option))
        return out

    def converter(self, section, option):
        schema = self.schema.get(section, option)
        call = None
        try:
            call = getattr(builtins, schema)
        except:
            try:
                call = find_name(schema)
            except:
                if len(schema) != 0:
                    def call(value):
                        match = re.search(schema, value)
                        if match is None:
                            raise ValueError
                        return match.group()
        return call

    def _transform_option(self, section, option, value):
        call = self.converter(section, option)

        if call is None:
            raise ValueError
        else:
            return call(value)



# Copyright (C) 2017 Jan Jancar
#
# This file is a part of the Mailman PGP plugin.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

""""""
import logging

from public import public

from mailman_pgp.config import config, mm_config
from mailman_pgp.database import transaction
from mailman_pgp.model.list import PGPMailingList
from mailman_pgp.pgp.keygen import ListKeyGenerator

log = logging.getLogger('mailman.plugin.pgp.styles')


@public
class PGPStyle:
    def apply(self, mailing_list):
        """Creates the encrypted mailing list instance for the list it's
        applied to.
        """
        pgp_list = PGPMailingList.for_list(mailing_list)
        if pgp_list:
            return

        generate = config.get_value('keypairs', 'autogenerate')

        mailing_list.posting_chain = 'pgp-posting-chain'

        old_policy = mailing_list.subscription_policy.name
        new_policy_name = 'pgp-' + old_policy[4:]
        if new_policy_name in mm_config.workflows:
            mailing_list.subscription_policy = new_policy_name

        with transaction() as session:
            log.debug(
                'Creating a PGP mailing list {}.'.format(mailing_list.list_id))
            pgp_list = PGPMailingList(mailing_list)
            if generate:
                keygen = ListKeyGenerator(pgp_list)
                keygen.generate()
            session.add(pgp_list)
